gameobject "event_director"
<<
	local
	<<

	>>

	state
	<<
		int updateTimer
		bool newPoolStartFired
		string choice
		string currentArcName
		string queuedArcName
		string name
		table eventArc
		table eventData
		table arcsCompleted
		int lastArcDay
		int arcStartCounter
		int arcPoints
		gameObjectHandle event_building
	>>

	receive Create( stringstringMapHandle init )
	<<
		if init.name then
			-- you're a special type of event_director
			printl("events", "event_director: received name " .. init.name)
			state.name = init.name
			send("gameSession","setSessiongOH", "event_director_" .. init.name , SELF) -- note underscore!
		else
			-- otherwise you're a generic event director
			state.name = ""
			send("gameSession","setSessiongOH", "event_director", SELF)
		end
		
		for k,v in pairs(EntityDB.eventArcTable.arcs) do
			printl("events", "found event arc: " .. v.name)	
		end
		state.waitingArcName = nil
		state.endingTimer = -1
		state.queuedArcName = nil
		state.currentArcName = nil
		state.updateTimer = -1
		state.arcsCompleted = {}
		state.lastArcDay = 0
		state.arc_start_counter = -1
	>>

	receive startNewEventArc( string arcname )
	<<
		if not arcname then
			printl("events", "received startNewEventArc with nil arcname!") 
			return
		end
		printl("events", "received startNewEventArc: " .. tostring( arcname) )

		if state.currentArcName then
			printl("events", "WARNING event_director " .. state.name .. " received startNewEventArc: " ..  tostring(arcname) .. " but was already executing " .. state.currentArcName )
			--printl("events", "aborting old event and beginning new one now, because we're assuming this is intentional")
			--send(SELF,"endEventArcImmediate")
			return
		end

		state.currentArcName = state.waitingArcName
		state.waitingArcName = nil
		
		state.eventData = {} -- use to store data throughout the entire arc of events
		state.eventData.eventPoints = 0 -- used for eventLocked things.
		state.event_building = nil
		
		state.currentArcName = arcname
		for k,v in pairs(EntityDB.eventArcTable.arcs) do
			if v.name == state.currentArcName then
				state.eventArc = v
				break
			end
		end
		
		state.currentArcStartDay = query("gameSession","getSessionInt","dayCount")[1]
		
		if state.eventArc.max_duration_days then
			state.currentArcEndDay = state.currentArcStartDay + state.eventArc.max_duration_days
		else
			state.currentArcEndDay = -1
		end
		
		local name = state.currentArcName
		if state.eventArc.hide_name then
			if state.name == "eldritch" then
				local names = {
					"An Abnormal Happening",
					"Worrying Events",
					"Untoward Events",
					"Strange Events",
					"Weird Happenings",
					"A Strange Occurance",
					"Mysterious Events",
					"The Eldritch Time",
					"Unexplainable Oddities",
					"An Unnatural Juncture",
					"A Case of the Supernatural",
					"Supernatural Rumours",
					"Dark Tidings",
					"Occult Events",
					"Occult Occurances",
					"Eldritch Events",
				}
				name = names[ rand(1,#names) ]
			else
				local names = {
					"Interesting Times",
					"Something Happening",
					"Goings-on",
					"A Notable Event",
				}
				name = names[ rand(1,#names) ]
			end
		end
		
		send("rendCommandManager",
			"odinRendererPersistentAlertMessage",
			name,		-- headingText
			false,					-- dismissable? (false)
			SELF)					-- self gameObject
		
		if state.eventArc.descriptionPast and
			state.eventArc.descriptionCurrent then
			
			send(SELF,
				"forwardEventStatusText",
				state.eventArc.descriptionPast,
				state.eventArc.descriptionCurrent)		
		end
		
		if state.eventArc.stages then
			send(SELF,"doStage", "start")
			
		elseif state.eventArc.eventName then
			-- if no stages given, just fire the event.
			state.currentEvent = query("gameSimEventManager",
								 "startEvent",
								 state.eventArc.eventName,
								 {},
								 { director_name = "_" .. state.name })[1]
			
			-- it will be assumed that the event fired will handle ending the arc w/ director.
		end
	>>
	
	receive doStage( string stagename )
	<<
		printl("events", "event director " .. tostring(state.name) .. " received stagename: " .. tostring(stagename) )
		-- 1. resolve stagename, if necessary.
		if not stagename then
			-- 1. find if progress condition exists
			if state.nextStageTrigger == "choice" or
				(state.nextStageTrigger == "timer" and state.eventData and state.eventData.choice) then
				
				printl("events", "got choice for next stage: " .. state.eventData.choice )
				
				-- find if choice was made use that, if not use default.
				if state.eventData and state.eventData.choice then
					if state.eventArc.stages[ state.currentStageName ].nextStage[ state.eventData.choice ] then
						stagename =  state.eventArc.stages[ state.currentStageName ].nextStage[ state.eventData.choice ]
					else
						stagename = state.eventArc.stages[ state.eventData.choice ]
					end
				end
				
			elseif state.nextStageTrigger == "pointsThreshhold" then
				if not state.eventData.points_threshold_achieved then
					-- failure, if it exists
					if state.eventArc.stages[ state.currentStageName ].nextStage.failure then
						stagename = state.eventArc.stages[ state.currentStageName ].nextStage.failure
					end
				end
				-- else will hit "default"
			end
			
			-- 2. or just use default (but if not default, go to "ending"=)
			if not stagename then
				if state.eventArc.stages[ state.currentStageName ].nextStage.default then
					stagename = state.eventArc.stages[ state.currentStageName ].nextStage.default
				else
					stagename = "ending"
				end
			end
			
			-- okay, now use stagename as key to pull actual stage name from list.
			printl("events", "event_director " .. tostring(state.name) .. " found stagename: " .. stagename )
		end
		-- else just use given eventname.
		
	-- 2. now clean out old stage data & put in new stage data
		state.currentStageName = stagename
		--state.stageProgressType = state.eventArc.stages[ stagename ].nextStageTrigger.trigger
		
		local worldstats = EntityDB.WorldStats
		local debugVar1 = state.eventArc.stages
		local tinfo = state.eventArc.stages[ stagename ].nextStageTrigger
		if tinfo then 
			if tinfo.trigger == "daysElapsed" then
				if tinfo.min and tinfo.max then
					state.updateTimer = rand(tinfo.min,tinfo.max) * worldstats.dayNightCycleTenthSeconds
				else
					state.updateTimer = tinfo.value * worldstats.dayNightCycleTenthSeconds
				end
				state.nextStageTrigger = "timer"
			elseif tinfo.trigger == "shiftsElapsed" then
				if tinfo.min and tinfo.max then
					state.updateTimer = rand(tinfo.min,tinfo.max) * div( worldstats.dayNightCycleTenthSeconds, worldstats.numberOfShiftsInDay)
				else
					state.updateTimer = tinfo.value * div( worldstats.dayNightCycleTenthSeconds, worldstats.numberOfShiftsInDay)
				end
				state.nextStageTrigger = "timer"
			elseif tinfo.trigger == "secondsElapsed" then
				if tinfo.min and tinfo.max then
					state.updateTimer = rand(tinfo.min,tinfo.max) * 10
				else
					state.updateTimer = tinfo.value * 10
				end
				state.nextStageTrigger = "timer"
				
			elseif tinfo.trigger == "choice" or
				tinfo.trigger == "manual" then
				
				state.nextStageTrigger = tinfo.trigger
				state.updateTimer = -1
				
			elseif tinfo.trigger == "pointsThreshhold" then
				if tinfo.min and tinfo.max then
					state.nextStagePointsThreshhold = rand(tinfo.min,tinfo.max)
				elseif tinfo.value then
					state.nextStagePointsThreshhold = tinfo.value
				else
					printl("events", "director: arc = " .. state.currentArcName.. " / attempted state.nextStagePointsThreshhold w/o value, defaulting to 1")
					state.nextStagePointsThreshhold = 1
				end
				
				send(SELF,"setKeyBool","points_threshold_achieved",false) -- reset this.
				state.eventData.eventPoints = 0
				state.nextStageTrigger = tinfo.trigger
				state.updateTimer = -1
			end
		else
			-- ending?
			printl("events", "no trigger info found for " .. stagename)
			state.updateTimer = -1
			state.nextStageTrigger = "manual"
		end
		printl("events", "trigger for next stage = " .. state.nextStageTrigger )
		state.eventData.choice = nil -- clear out old choice string
	
	-- 3. and launch new stage's event
		
		if state.eventArc.stages[state.currentStageName] then
			state.currentEvent = query("gameSimEventManager",
								 "startEvent",
								 state.eventArc.stages[state.currentStageName].eventName,
								 {},
								 { director_name = "_" .. state.name })[1]
			
		-- 4. and update description strings, if available
		
			if state.eventArc.stages[state.currentStageName].statusPast and
				state.eventArc.stages[state.currentStageName].statusCurrent then
				
				send(SELF,"forwardEventStatusText",
					state.eventArc.stages[state.currentStageName].statusPast,
					state.eventArc.stages[state.currentStageName].statusCurrent)
			end
		else
			printl("events", state.name .. " found no stage entry for " .. tostring(state.currentStageName) .. "! hopefully this is intended.")
		end
	>>
	
	receive Update()
	<<
		if state.endingTimer then
			if state.endingTimer > 0 then
				state.endingTimer = state.endingTimer - 1
				if state.endingTimer == 0 then
					send(SELF,"endEventArcImmediate")
				end
				return
			end
		end
	
		if state.nextStageTrigger == "timer" and
			state.updateTimer >= 0 then

			if state.updateTimer == 0 then
			
				printl("events", " updatetimer hit zero, doing next event if possible ")
				
				-- first check if our entire arc is not overtime for the finale (if so, trigger end)
				if state.eventArc.stages[ state.currentStageName ].max_duration_days and
					query("gameSession","getSessionInt","dayCount")[1] >= state.currentArcEndDay then
					
					printl("events", " hit event arc time limit, doing end event" )
					state.currentEvent = query("gameSimEventManager",
												"startEvent",
												 state.eventArc.stages.ending.eventName, 
												{},
												{ director_name = state.name })[1]
					
				else
					-- otherwise time for next event!
					printl("events", "   doing next event in valid pool as planned ")
					send(SELF,"doStage", nil)
				end
			end
			
			state.updateTimer = state.updateTimer - 1
			
		elseif state.nextStageTrigger == "choice" and state.eventData.choice ~= nil then			
			send(SELF,"doStage",nil)
			
		elseif state.nextStageTrigger == "pointsThreshhold" and state.eventData.eventPoints >= state.nextStagePointsThreshhold then
			
			printl("events", "director reached pointsThreshhold of: " .. state.nextStagePointsThreshhold )
			
			send(SELF,"setKeyBool","points_threshold_achieved",true)
			
			state.eventData.eventPoints = nil
			state.nextStagePointsThreshhold = nil
			
			send(SELF,"doStage",nil)
		end
		
		if state.arc_start_counter > 0 then
			state.arc_start_counter = state.arc_start_counter -1
			if state.arc_start_counter == 0 then
				state.arc_start_counter = -1
				send(SELF,"startNewEventArc",state.waitingArcName)
			end
		end
	>>

	receive Nightfall()
	<<

	>>
	
	receive Sunrise()
	<<
		if state.name == "_tutorial" or
			state.name == "_other" or
			state.name == "other" or
			state.name == "tutorial" then
			
			-- tutorial direction should not do this.
			return
		end
		
		--printl("events", state.name .. " event director received Sunrise.")
		
		if not state.currentArcName then
			send(SELF,"resolveArcPoints")
		end
	>>
	
	receive ShiftUpdate( int shiftnumber)
	<<
	
	>>
	
	receive setKeyInt( string key, int i)
	<<
		printl("events", "event_director " .. tostring(state.name) .. ": got setKeyInt: " .. key .. " : " .. i)
		state.eventData[key] = i
	>>
	
	respond getKeyInt( string key )
	<<
		return "getKeyIntMessage", state.eventData[key]
	>>
	
	receive setKeyBool( string key, bool b)
	<<
		printl("events", "event_director: got setKeyBool: " .. key .. " : " .. tostring(b))
		state.eventData[key] = b
	>>
	
	respond getKeyBool( string key )
	<<
		return "getKeyBoolMessage", state.eventData[key]
	>>
	
	receive setKeyString( string key, string s)
	<<
		printl("events", "event_director: got setKeyString: " .. key .. " : " .. tostring(s))
		state.eventData[key] = s
	>>
	
	receive incKeyInt( string key, int i)
	<<
		printl("events", "event_director: got incKeyInt: " .. key .. " : " .. i)
		if not state.eventData[key] then
			state.eventData[key] = i
			printl("events", "incKeyInt is now: " .. state.eventData[key] )
		else
			state.eventData[key] = state.eventData[key] + i
			printl("events", "incKeyInt is now: " .. state.eventData[key] )
		end
	>>
	
	receive setKeyGameobject( string key, gameObjectHandle gOH)
	<<
		printl("events", "event_director: got gOH for key: " .. key .. "; gOH = " .. tostring(gOH) )
		state.eventData[key] = gOH
	>>
	
	respond getKeyString( string key )
	<<
		return "getKeyStringMessage", state.eventData[key]
	>>
	
	respond getKeyInt( string key )
	<<
		return "getKeyIntMessage", state.eventData[key]
	>>
	
	respond getKeyGameobject( string key )
	<<
		return "getKeygOHMessage", state.eventData[key]
	>>
	
	receive arcAbortDueToRequirements()
	<<
		if state.currentArcName then
			printl("events", "event_director " .. tostring(state.name) .. " got arcAbortDueToRequirements while attempting: " .. tostring(state.currentArcName) )
			-- aborts the arc without recording it as completed.
			-- use this early in the start event of an arc to abort if the arc cannot be done properly for some reason.
			
			--send(SELF,"forwardEventStatusText","","")
			
			state.lastArcDay = query("gameSession","getSessionInt","dayCount")[1]
			--state.arcsCompleted[ state.currentArcName ] = state.lastArcDay -- store day this arc was completed
			
			-- refund arc points
			if state.eventArc.cost then
				local cost = state.eventArc.cost
				state.arcPoints = state.arcPoints + cost
			end
		else
			printl("events", "event_director " .. tostring(state.name) .. " got arcAbortDueToRequirements with no arc active.")
		end
			
		state.eventData = {}
		state.nextStageTrigger = nil
		state.updateTimer = -1
		state.currentArcName = nil
		state.currentArcEndDay = nil
		state.currentArcStartDay = nil
		state.currentEvent = nil
		
		send("rendCommandManager","odinRendererRemovePersistentEventMessage",SELF)
		
		send("rendCommandManager","odinRendererForceAlertUpdate")
	>>
	
	receive endEventArc()
	<<
		printl("events", "event_director " .. tostring(state.name) .. " got endEventArc while doing " .. tostring(state.currentArcName) .. "; ending timer set.")
		state.endingTimer = 150
	>>
	
	receive endEventArcImmediate()
	<<
		printl("events", "event_director " .. tostring(state.name) .. " got endEventArc while doing " .. tostring(state.currentArcName) )

		state.lastArcDay = query("gameSession","getSessionInt","dayCount")[1]
		if state.currentArcName then
			state.arcsCompleted[ state.currentArcName ] = state.lastArcDay -- store day this arc was completed
		end
		
		state.eventData = {}
		state.nextStageTrigger = nil
		state.updateTimer = -1
		state.currentArcName = nil
		state.currentArcEndDay = nil
		state.currentArcStartDay = nil
		state.currentEvent = nil
		
		send("rendCommandManager","odinRendererRemovePersistentEventMessage",SELF)
		
		send("rendCommandManager","odinRendererForceAlertUpdate")
		
		printl("events", "arc ended.")
	>>
	
	receive tryToStartEventArc( string arcname )
	<<
		printl("events", state.name .. " received tryToStartEventArc: " .. arcname )
		-- if no event arc is active, try to start the named event arc.
		if not state.currentArcName then
			if EntityDB[ "eventArcTable" ].arcs[ arcname ] then
				send(state.director,"startNewEventArc",arcname)
			else
				printl("events", "WARNING: " .. state.name .. " attempted tryToStartEventArc with: " .. arcname .. ", but could not find that event in db!")
			end
		else
			printl("events", state.name .. " attempted tryToStartEventArc with: " .. arcname .. ", but another event was active")
		end
	>>

	receive forwardEventStatusText(string past, string present)
	<<
		send("rendCommandManager",
			"odinRendererUpdateEventStatusText",
			past,
			present,
			SELF)
		
		if state.eventArc then
			send("rendCommandManager",
				"odinRendererTickerMessage",
				"Day " .. query("gameSession","getSessionInt","dayCount")[1] .. ": " .. past,
				state.eventArc.icon,
				state.eventArc.iconsheet)
			
		end
	>>
	
	respond getName()
	<<
		return "nameResponse", "event_director_" .. state.name
	>>
	
	respond getEventArcName()
	<<
		return "getEventArcNameResponse", state.currentArcName
	>>
	
	receive releaseEventBuilding()
	<<
		printl("events", "director " .. state.name .. " got releaseEventBuilding")
		if state.eventData["pointsBuilding"] then
			
			send(state.eventData["pointsBuilding"],"releaseEventBuilding")
			state.eventData["pointsBuilding"] = nil
			
		end
		
		if state.nextStageTrigger == "pointsThreshhold" then
			if not state.eventData.points_threshold_achieved then
				send(SELF,"setKeyBool","points_threshold_achieved",false)
				send(SELF,"setKeyBool","eventBuildingLost",true)
				
				-- ??? Gotta move the story along.
				-- you CAN set a fallback stage if you need one.
			if not state.eventData.eventBuildingLostFallbackStage then
					--send(SELF,"doStage","ending")
					send(SELF,"doStage",nil)
				else
				send(SELF,"doStage",state.eventData.eventBuildingLostFallbackStage)
				end
			end
		elseif state.nextStageTrigger == "daysElapsed" or
			state.nextStageTrigger == "shiftsElapsed" or
			state.nextStageTrigger == "secondsElapsed" then
			
			-- on timer, but building is released.
			if query(SELF,"getKeyBool","eventBuildingOwnerNil")[1] == true then
				-- this might be a problem.force next stage.
				send(SELF,"doStage",nil)
			end
			--send(SELF,"setKeyBool","eventBuildingWorkInterrupted"

		end
	>>
	
	respond getEventBuilding()
	<<
		local building = nil
		if state.eventData["pointsBuilding"] then
			building = state.eventData["pointsBuilding"]
		end
		return "getEventBuildingMessage", building
	>>
	
	receive killDirector()
	<<
		printl("events", "director " .. state.name .. " got killDirector")
		send(SELF,"DestroySelf", nil )
	>>
	
	respond getEventBuildingArt()
	<<
		local art = "ui\\eventart\\doing_science.png"
		if state.eventData["eventBuildingArt"] then
			art = state.eventData["eventBuildingArt"]
		end
		return "getEventBuildingArtMessage",art
	>>
	
	respond getEventBuildingText()
	<<
		local text = "Work on a special project is being carried out."
		if state.eventData["eventBuildingText"] then
			text = state.eventData["eventBuildingText"]
		end
		return "getEventBuildingTextMessage",text
	>>
	
	receive addArcPoints( int i )
	<<
		printl("events", tostring(state.name) .. " got addArcPoints: " .. i  )
		state.arcPoints = state.arcPoints + i
		send("gameSession","incSessionInt","disturbancePoints",i)
	>>
	
	receive queueNextArc( string name )
	<<
		printl("events", state.name .. " event director received queueNextArc: " .. tostring(name) )
		
		if name then
			state.queuedArcName = name
			printl("events", state.name .. " event director has queued: " .. tostring(name) )
		else
			-- find something to add to the queue!
			local today = query("gameSession","getSessionInt","dayCount")[1]

			local arcPool = EntityDB.eventArcTable.arcs
			local arcPoolFinal = {}
			
			for k,v in pairs(arcPool) do
				local valid = true
					
				-- can't queue same arc in 20 day period.
				if state.arcsCompleted[ v.name ] then
					if 20 < (today - state.arcsCompleted[ v.name ]) then
						valid = false
					end	
				end
				
				if not v.minDayCount then
					-- implicitly valid for this case.
				elseif v.minDayCount and today < v.minDayCount then
					valid = false
				end

				if v.filter == "debug" or v.filter == "tutorial" then
					valid = false

				elseif state.name and
					state.name ~= "" and
					v.filter ~= state.name then
					
					valid = false
				end
				
				if valid == true then
					if v.weight then
						for i=1, v.weight do
							table.insert(arcPoolFinal,v)
						end
					else
						table.insert(arcPoolFinal,v)
					end
				end
			end
			
			for k,v in pairs(arcPoolFinal) do
				printl("DAVID", state.name .. " : " .. tostring(k) .. " - " .. tostring(v.name) )	
			end
			
			if #arcPoolFinal == 0 then
				printl("events", state.name .. " event director doing queueNextArc found no valid arc to queue." )
				return
			end
				
			local key = rand(1,#arcPoolFinal)
			
			state.queuedArcName = arcPoolFinal[key].name
			printl("events", state.name .. " event director has queued: " .. state.queuedArcName )
		end
	>>
	
	receive resolveArcPoints()
	<<
		if state.currentArcName then
			printl("events", state.name .. " event director received resolveArcPoints but is doing arc, aborting." )
			return
		end
		
		printl("events", state.name .. " event director doing resolveArcPoints, points to resolve with = " .. tostring(state.arcPoints) )
		-- do we have a next arc queued?
		-- if not, find one
		if not state.queuedArcName then
			send(SELF,"queueNextArc",nil)
			if not state.queuedArcName then
				-- really didn't find out? Then give up.
				printl("events", state.name .. " event director tried to resolveArcPoints but found no valid arc")
				return
			end
		end
		
		local arcData = false
		
		for k,v in pairs(EntityDB.eventArcTable.arcs) do
			if v.name == state.queuedArcName then
				arcData = v
				break
			end
		end
		
		local cost = arcData.cost
		
		printl("events", state.name .. " event director testing points (" .. state.arcPoints .. ") vs. cost (".. cost .. ") of arc: " .. state.queuedArcName )
		
		if cost <= state.arcPoints then
			state.arcPoints = state.arcPoints - cost
			send("gameSession","incSessionInt","disturbancePoints",cost * -1)
			
			if state.arcPoints > 0 then
				send("gameSession","incSessionInt","arcPointsPool",state.arcPoints)
				state.arcPoints = 0
			end
			
			-- set arc to start!
			state.waitingArcName = state.queuedArcName
			state.queuedArcName = nil
			
			-- if starting shift, then set counter to that.
			if arcData.startAtShift then
				state.arc_start_counter = (arcData.startAtShift -1) * EntityDB.WorldStats.shiftLengthTenthSeconds
			else
				state.arc_start_counter = rand(1,EntityDB.WorldStats.dayNightCycleTenthSeconds)
			end
		end
	>>
>>
